import asyncdispatch, base64, jester, json, jwt, logging, options, sequtils,
  strutils, times, uuids
import timeutils except `<`

import ./db, ./configuration, ./models, ./service, ./version

const JSON = "application/json"

type
  Session* = object
    user*: User
    issuedAt*, expires*: Time

proc newSession*(user: User): Session =
  result = Session(
    user: user,
    issuedAt: getTime().utc.trimNanoSec.toTime,
    expires: daysForward(1).trimNanoSec.toTime)

template halt(code: HttpCode,
              headers: RawHeaders,
              content: string): typed =
  ## Immediately replies with the specified request. This means any further
  ## code will not be executed after calling this template in the current
  ## route.
  bind TCActionSend, newHttpHeaders
  result[0] = CallbackAction.TCActionSend
  result[1] = code
  result[2] = some(headers)
  result[3] = content
  result.matched = true
  break allRoutes

template jsonResp(code: HttpCode, details: string = "", headers: RawHeaders = @{:} ) =
  halt(
    code,
    headers & @{"Content-Type": JSON},
    $(%* {
      "statusCode": code.int,
      "status": $code,
      "details": details
    })
  )

template json500Resp(ex: ref Exception, details: string = ""): void =
  when not defined(release): debug ex.getStackTrace()
  error details & ":\n" & ex.msg
  jsonResp(Http500)

proc toJWT*(ctx: PMApiContext, session: Session): string =
  ## Make a JST token for this session.
  var jwt = JWT(
    header: JOSEHeader(alg: HS256, typ: "jwt"),
    claims: toClaims(%*{
      "sub": $(session.user.id),
      "iat": session.issuedAt.toUnix.int,
      "exp": session.expires.toUnix.int }))

  jwt.sign(ctx.cfg.authSecret)
  result = $jwt

proc fromJWT*(ctx: PMApiContext, strTok: string): Session =
  ## Validate a given JWT and extract the session data.
  let jwt = toJWT(strTok)
  var secret = ctx.cfg.authSecret
  if not jwt.verify(secret): raiseEx "Unable to verify auth token."
  jwt.verifyTimeClaims()

  # Find the user record (if authenticated)
  let userId = jwt.claims["sub"].node.str
  var user: User
  try: user = ctx.db.getUser(parseUUID(userId))
  except: raiseEx "unknown user"

  result = Session(
    user: user,
    issuedAt: fromUnix(jwt.claims["iat"].node.num),
    expires: fromUnix(jwt.claims["exp"].node.num))

proc fromApiToken*(ctx: PMApiContext, strTok: string): Session =
  info "fromApiToken"
  let pairs = strTok.decode.split(":")
  let email = pairs[0]
  let apiTokVal = pairs[1]

  info "email: " & email & "\tapiTokVal: " & apiTokVal
  # Look up the user
  var user: User
  try:
    let users = ctx.db.getUsersByEmail(email)
    if users.len != 1: raiseEx ""
    user = users[0]
    info "Found user: " & $user
  except: raiseEx "invalid username or password"

  # look for the ApiToken record, hashing the token provided with the user's salt
  let hashedToken = hashPwd(apiTokVal, user.salt)
  echo "Hashed token: " & hashedToken
  let foundTokens = ctx.db.getApiTokensByHashedToken(hashedToken)
  if foundTokens.len != 1: raiseEx "invalid username or password"

  let apiToken = foundTokens[0]

  # make sure that if we found the token it is for this user (don't expect
  # another user's salt to be able to hash this user's pwd correctly, but eh)
  if apiToken.userId != user.id: raiseEx "invalid username or password"

  # make sure the token has not expired
  if apiToken.expires.isSome and apiToken.expires.get < getTime().utc:
    raiseEx "invalid username or password"

  # finally, if we're here, we know that this is a valid api token
  result = Session(
    user: user,
    issuedAt: getTime().utc.trimNanoSec.toTime,
    expires: (getTime().utc.trimNanoSec + 1.minutes).toTime)

proc extractSession(ctx: PMApiContext, request: Request): Session =
  ## Helper to extract a session from a reqest.

  # Find the auth header
  if not request.headers.hasKey("Authorization"):
    raiseEx "No auth token."

  let headerVal = request.headers["Authorization"]

  # If they gave us an API Token, validate that to get the session
  if headerVal.startsWith("Basic "):
    result = fromApiToken(ctx, headerVal[6..^1])
  elif headerVal.startsWith("Bearer "):
    result = fromJWT(ctx, headerVal[7..^1])
  else:
    raiseEx "Invalid Authentication type ('Bearer' and 'Basic'  supported)."

proc makeAuthToken*(ctx: PMApiContext, email, pwd: string): string =
  ## Given a user's email and pwd, validate the combination and generate a JWT
  ## token string.

  if email.len == 0 or pwd.len == 0:
    raiseEx "fields 'username' and 'password' required"

  # find the user record
  var user: User
  try:
    let users = ctx.db.getUsersByEmail(email)
    if users.len != 1: raiseEx ""
    user = users[0]
  except: raiseEx "invalid username or password"

  if not validatePwd(user, pwd): raiseEx "invalid username or password"

  let session = newSession(user)

  result = toJWT(ctx, session)

template checkAuth() =
  ## Check this request for authentication and authorization information.
  ## Injects the session into the running context. If the request is not
  ## authorized, this template returns an appropriate 401 response.

  var session {.inject.}: Session

  try: session = extractSession(ctx, request)
  except:
    debug "Auth failed: " & getCurrentExceptionMsg()
    jsonResp(Http401, "Unauthorized", @{"WWW-Authenticate": "Bearer"})

proc start*(ctx: PMApiContext): void =

  if ctx.cfg.debug: setLogFilter(lvlDebug)

  var stopFuture = newFuture[void]()

  settings:
    port = Port(ctx.cfg.port)
    appName = "/api"

  routes:

    get "/version":
      resp($(%("personal_measure_api v" & PM_API_VERSION)), JSON)

    post "/auth-token":
      var email, pwd: string
      try:
        let jsonBody = parseJson(request.body)
        email = jsonBody["email"].getStr
        pwd = jsonBody["password"].getStr
      except: jsonResp(Http400)

      try:
        let authToken = makeAuthToken(ctx, email, pwd)
        resp($(%authToken), JSON)
      except: jsonResp(Http401, getCurrentExceptionMsg())

    get "/user":
      checkAuth()

      resp(Http200, $(%session.user), JSON)

    get "/api-tokens":
      checkAuth()

      resp(Http200, $(%ctx.db.getApiTokensByUserId($session.user.id)))

    post "/api-tokens":
      checkAuth()

      var newToken: ApiToken
      try:
        let jsonBody = parseJson(request.body)
        newToken = ApiToken(
          userId: session.user.id,
          name: jsonBody["name"].getStr,
          expires: none[DateTime](),
          hashedToken:  "")
      except: jsonResp(Http400)

      try:
        let tokenValue = randomString(ctx.cfg.pwdCost)
        newToken.hashedToken = hashPwd(tokenValue, session.user.salt)
        newToken = ctx.db.createApiToken(newToken)
        let respToken = %newToken
        respToken["value"] = %tokenValue
        resp($respToken, JSON)
      except:
        debug getCurrentExceptionMsg()
        jsonResp(Http500)

    delete "/api-tokens/@tokenId":
      info "Request to delete API Token"
      checkAuth()

      try:
        let token = ctx.db.getApiToken(parseUUID(@"tokenId"))
        if ctx.db.deleteApiToken(token): jsonResp(Http200)
        else: jsonResp(Http500)
      except: jsonResp(Http404)

    post "/service/debug/stop":
      if not ctx.cfg.debug: jsonResp(Http404)
      else:
        let shutdownFut = sleepAsync(100)
        shutdownFut.callback = proc(): void = complete(stopFuture)
        resp($(%"shutting down"), JSON)

  waitFor(stopFuture)

